/* schierke.c - Don Yang (uguu.org)

   10/30/11
                                                                           */

#include<time.h>
#include<stdlib.h>
#include"SDL.h"

#define WIDTH 800
#define HEIGHT 600
#define TRAIL_SIZE 64
#define ELEMENT_COUNT 4
#define SWARM_SIZE 256
#define FRAME_DELAY_MS (1000/60)

/* All particles, mapping from cycle to element to swarm of points to (x, y).
   At each frame, we update points in the next cycle, and raw particles from
   the previous cycle at a slightly dimmer color.

   Each value is a fixed point value, with higher 16 bits being the screen
   coordinates, and the lower 16 bits to be the fractional part.

   All swarm particles with point index > 0 follows the 0th point.           */
static Sint32 Position[TRAIL_SIZE][ELEMENT_COUNT][SWARM_SIZE][2];
static Sint32 Direction[ELEMENT_COUNT][SWARM_SIZE][2];
static int Cycle = 0;

/* Targets for swarm point 0, same fixed point format as above */
static Sint32 Target[ELEMENT_COUNT][2];

/* Shared indices */
static Sint32 i, j, k, element;

/* Drawing state */
Uint32 x, y, pixel;
SDL_Event event;
int run_loop = 1;
Uint32 current_time, next_time;
const SDL_VideoInfo *video_info;
SDL_Surface *screen;
Uint32 flags = SDL_HWSURFACE|SDL_DOUBLEBUF;
int bpp;

/* Color palette for each elemental */
static Uint8 Palette[ELEMENT_COUNT][3] =
{
   {255, 16, 16},
   {0, 255, 0},
   {255, 255, 64},
   {64, 128, 255}
};

/* Return a random number in the range of [min, max) */
static Sint32 Rand(Sint32 min, Sint32 max)
{
   return (max - min) * (double)(rand()) / (double)RAND_MAX + min;
}

/* Clamp value to a particular range */
static Sint32 Clamp(Sint32 value)
{
   return value < -k ? -k : (value > k ? k : value);
}

/* Update velocity for a single point to approach a target point */
static void Approach(Sint32 index, Sint32 tx, Sint32 ty,
                     Sint32 minv, Sint32 maxv)
{
   if( Position[Cycle][element][index][0] < tx )
      Direction[element][index][0] += Rand(minv, maxv);
   if( Position[Cycle][element][index][0] > tx )
      Direction[element][index][0] -= Rand(minv, maxv);
   if( Position[Cycle][element][index][1] < ty )
      Direction[element][index][1] += Rand(minv, maxv);
   if( Position[Cycle][element][index][1] > ty )
      Direction[element][index][1] -= Rand(minv, maxv);
   for(j = 0; j < 2; j++)
      Direction[element][index][j] = Clamp(Direction[element][index][j]);
}

/* Render a single frame */
#define RENDER(type) \
{                                                                            \
   /* At each cycle, we will update points such that cycle points at the     \
      latest set of points.  Since we want to draw points from oldest        \
      cycle to newest cycle, we will start at cycle+1.                   */  \
   for(i = 1; i <= TRAIL_SIZE; i++)                                          \
   {                                                                         \
      j = (Cycle + i) % TRAIL_SIZE;                                          \
      for(element = 0; element < ELEMENT_COUNT; element++)                   \
      {                                                                      \
         /* The leader point will have shades of white */                    \
         pixel = SDL_MapRGB(screen->format,                                  \
                            i * 255 / TRAIL_SIZE,                            \
                            i * 255 / TRAIL_SIZE,                            \
                            i * 255 / TRAIL_SIZE);                           \
         for(k = 0; k < SWARM_SIZE; k++)                                     \
         {                                                                   \
            /* After drawing the leader point, set color to draw             \
               follower points.                                  */          \
            if( k == 1 )                                                     \
            {                                                                \
               pixel = SDL_MapRGB(screen->format,                            \
                                  Palette[element][0] * i / TRAIL_SIZE,      \
                                  Palette[element][1] * i / TRAIL_SIZE,      \
                                  Palette[element][2] * i / TRAIL_SIZE);     \
            }                                                                \
                                                                             \
            /* Scale coordinates and clip */                                 \
            x = Position[j][element][k][0];                                  \
            x >>= 16;                                                        \
            if( x < 0 || x >= WIDTH )                                        \
               continue;                                                     \
            y = Position[j][element][k][1];                                  \
            y >>= 16;                                                        \
            if( y < 0 || y >= HEIGHT )                                       \
               continue;                                                     \
                                                                             \
            /* Draw pixel */                                                 \
            *(type*)((Uint8*)(screen->pixels) +                              \
                              y * screen->pitch +                            \
                              x * bpp)                                       \
               = (type)pixel;                                                \
         }                                                                   \
      }                                                                      \
   }                                                                         \
}

static void Render16() RENDER(Uint16)
static void Render32() RENDER(Uint32)

int main(int argc, char **argv)
{
   /* Initialize SDL */
   if( SDL_Init(SDL_INIT_VIDEO) < 0 )
   {
      printf("Unable to init SDL: %s\n", SDL_GetError());
      return 1;
   }
   atexit(SDL_Quit);

   /* Initialize surface */
   video_info = SDL_GetVideoInfo();
   bpp = video_info->vfmt->BytesPerPixel;
   if( bpp != 2 && bpp != 4 )
   {
      puts("Unsupported pixel format");
      SDL_Quit();
      return 1;
   }
   if( argc > 1 )
      flags |= SDL_FULLSCREEN;
   screen = SDL_SetVideoMode(WIDTH, HEIGHT, bpp * 8, flags);
   if( screen == NULL )
   {
      printf("Error initializing screen: %s\n", SDL_GetError());
      SDL_Quit();
      return 1;
   }
   SDL_WM_SetCaption("Schierke", NULL);
   SDL_ShowCursor(SDL_DISABLE);

   /* Initialize random seed */
   srand(time(NULL));

   /* Initialize swarms */
   for(element = 0; element < ELEMENT_COUNT; element++)
   {
      for(i = 0; i < SWARM_SIZE; i++)
      {
         Position[0][element][i][0] = Rand(0, WIDTH * 0x10000);
         Position[0][element][i][1] = Rand(0, HEIGHT * 0x10000);
         for(j = 0; j < 2; j++)
            Direction[element][i][j] = Rand(-0x30000, 0x30000);
      }
      for(i = 1; i < TRAIL_SIZE; i++)
      {
         for(j = 0; j < SWARM_SIZE; j++)
         {
            for(k = 0; k < 2; k++)
               Position[i][element][j][k] = Position[0][element][j][k];
         }
      }
      Target[element][0] = Rand(0, WIDTH * 0x10000);
      Target[element][1] = Rand(0, HEIGHT * 0x10000);
   }

   next_time = SDL_GetTicks() + FRAME_DELAY_MS;
   while( run_loop )
   {
      /* Render points */
      SDL_LockSurface(screen);
      SDL_FillRect(screen, NULL, SDL_MapRGB(screen->format, 0, 0, 0));
      if( screen->format->BytesPerPixel == 4 )
         Render32();
      else
         Render16();
      SDL_UnlockSurface(screen);
      SDL_Flip(screen);

      /* Update animation state */
      for(element = 0; element < ELEMENT_COUNT; element++)
      {
         /* Update target for leader about once per second */
         if( Rand(0, 60) == 0 )
         {
            Target[element][0] = Rand(0x10000 * (WIDTH / 10),
                                      0x10000 * (WIDTH - (WIDTH / 10)));
            Target[element][1] = Rand(0x10000 * (HEIGHT / 10),
                                      0x10000 * (HEIGHT - (HEIGHT / 10)));
         }
         k = 0x20000;
         Approach(0,
                  Target[element][0], Target[element][1],
                  0x10, 0x1000);

         /* Update velocities for followers */
         k <<= 1;
         for(i = 1; i < SWARM_SIZE; i++)
         {
            Approach(i,
                     Position[Cycle][element][0][0],
                     Position[Cycle][element][0][1],
                     0x20, 0x1800);
         }

         /* Update point positions */
         k = (Cycle + 1) % TRAIL_SIZE;
         for(i = 0; i < SWARM_SIZE; i++)
         {
            for(j = 0; j < 2; j++)
               Position[k][element][i][j] =
                  Position[Cycle][element][i][j] + Direction[element][i][j];
         }
      }
      Cycle = (Cycle + 1) % TRAIL_SIZE;

      /* Check for keyboard events */
      while( SDL_PollEvent(&event) )
      {
         if( event.type == SDL_KEYDOWN || event.type == SDL_QUIT )
            run_loop = 0;
      }

      /* Sleep until next frame */
      current_time = SDL_GetTicks();
      if( current_time < next_time )
         SDL_Delay(next_time - current_time);
      next_time += FRAME_DELAY_MS;
   }

   /* Cleanup */
   SDL_FreeSurface(screen);
   SDL_Quit();
   return 0;
}
